@@ -19,11 +19,12 @@
 Methods to annotate the Abstract Syntax Tree (AST) of an IDP-Z3 program.
 
 """
+from __future__ import annotations
 
 from copy import copy, deepcopy
 from itertools import chain
 import string
-from typing import Dict
+from typing import Dict, List
 
 from .Parse import (Vocabulary, Import, TypeDeclaration, Declaration,
     SymbolDeclaration, VarDeclaration, TheoryBlock, Definition, Rule,
@@ -34,11 +35,11 @@ from .Expression import (Expression, Symbol, SYMBOL, Type, TYPE,
     ARImplication, AImplication, AEquivalence,
     Operator, AComparison, AUnary, AAggregate,
     AppliedSymbol, UnappliedSymbol, Variable, VARIABLE, Brackets,
-    FALSE, SymbolExpr, Number, NOT, EQUALS, AND, OR, FALSE,
+    FALSE, SymbolExpr, Number, NOT, EQUALS, AND, OR, TRUE, FALSE,
     IMPLIES, RIMPLIES, EQUIV, FORALL, EXISTS, Extension)
 
 from .utils import (BOOL, INT, REAL, DATE, CONCEPT, RESERVED_SYMBOLS,
-    OrderedSet, IDPZ3Error, DEF_SEMANTICS, Semantics)
+    OrderedSet, IDPZ3Error, Semantics)
 
 
 # Class Vocabulary  #######################################################
@@ -153,7 +154,7 @@ Symbol.annotate = annotate
 
 # Class Type  #######################################################
 
-def annotate(self, voc, q_vars={}):
+def annotate(self, voc, q_vars):
     Symbol.annotate(self, voc, q_vars)
     if self.out:
         self.ins = [s.annotate(voc, q_vars) for s in self.ins]
@@ -213,7 +214,8 @@ Definition.annotate = annotate
 def get_instantiables(self,
                       interpretations: Dict[str, SymbolInterpretation],
                       extensions: Dict[str, Extension],
-                      for_explain=False):
+                      for_explain=False
+                      ) -> Dict[SymbolDeclaration, List[Expression]]:
     """ compute Definition.instantiables, with level-mapping if definition is inductive
 
     Uses implications instead of equivalence if `for_explain` is True
@@ -229,12 +231,13 @@ def get_instantiables(self,
     result = {}
     for decl, rules in self.canonicals.items():
         rule = rules[0]
-        rule.is_whole_domain = all(s.extension(interpretations, extensions)[0] is not None
+        rule.has_finite_domain = all(s.extension(interpretations, extensions)[0] is not None
                                    for s in rule.definiendum.decl.sorts)
-        if not rule.is_whole_domain:
-            self.check(rule.definiendum.symbol.decl not in self.level_symbols,
-                       f"Cannot have inductive definitions on infinite domain")
-        else:
+        inductive = (self.mode != Semantics.COMPLETION
+            and rule.definiendum.symbol.decl in self.level_symbols)
+
+        if rule.has_finite_domain or inductive:
+            # add a constraint containing the definition over the full domain
             if rule.out:
                 expr = AppliedSymbol.make(rule.definiendum.symbol,
                                           rule.definiendum.sub_exprs[:-1])
@@ -245,23 +248,20 @@ def get_instantiables(self,
                                           rule.definiendum.sub_exprs)
                 head.in_head = True
 
-            inductive = (not rule.out and DEF_SEMANTICS != Semantics.COMPLETION
-                and rule.definiendum.symbol.decl in rule.parent.level_symbols)
-
             # determine reverse implications, if any
             bodies, out = [], []
             for r in rules:
                 if not inductive:
                     bodies.append(r.body)
                     if for_explain and 1 < len(rules):  # not simplified -> no need to make copies
-                        out.append(RIMPLIES([head, r.body], r.annotations))
+                        out.append(IMPLIES([r.body, head], r.annotations))
                 else:
                     new = r.body.split_equivalences()
                     bodies.append(new)
                     if for_explain:
-                        new = deepcopy(new).add_level_mapping(rule.parent.level_symbols,
-                                             rule.definiendum, False, False)
-                        out.append(RIMPLIES([head, new], r.annotations))
+                        new = deepcopy(new).add_level_mapping(self.level_symbols,
+                                             rule.definiendum, False, False, self.mode)
+                        out.append(IMPLIES([new, head], r.annotations))
 
             all_bodies = OR(bodies)
             if not inductive:
@@ -271,13 +271,15 @@ def get_instantiables(self,
                     out = [EQUIV([head, all_bodies], self.annotations)]
             else:
                 if not out:  # no reverse implication yet
-                    new = deepcopy(all_bodies).add_level_mapping(rule.parent.level_symbols,
-                                             rule.definiendum, False, False)
-                    out = [RIMPLIES([deepcopy(head), new], self.annotations)]
-                all_bodies = deepcopy(all_bodies).add_level_mapping(rule.parent.level_symbols,
-                                        rule.definiendum, True, True)
+                    new = deepcopy(all_bodies).add_level_mapping(self.level_symbols,
+                                             rule.definiendum, False, False, self.mode)
+                    out = [IMPLIES([new, deepcopy(head)], self.annotations)]
+                all_bodies = deepcopy(all_bodies).add_level_mapping(self.level_symbols,
+                                        rule.definiendum, True, True, self.mode)
                 out.append(IMPLIES([head, all_bodies], self.annotations))
             result[decl] = out
+        else:
+            out = TRUE
     return result
 Definition.get_instantiables = get_instantiables
 
@@ -329,10 +331,10 @@ def rename_args(self, new_vars):
             self.body = AND([eq, self.body])
 
     for v in vars.values():
-        self.body = EXISTS([Quantee.make(v, v.sort)], self.body)
+        self.body = EXISTS([Quantee.make(v, sort=v.sort)], self.body)
 
     self.definiendum.sub_exprs = list(new_vars.values())
-    self.quantees = [Quantee.make(v, v.sort) for v in new_vars.values()]
+    self.quantees = [Quantee.make(v, sort=v.sort) for v in new_vars.values()]
     return self
 Rule.rename_args = rename_args
 
@@ -536,13 +538,8 @@ Expression.annotate = annotate
 def annotate1(self):
     " annotations that are common to __init__ and make() "
     self.variables = set()
-    if self.value is not None:
-        pass
-    if self.simpler is not None:
-        self.variables = self.simpler.variables
-    else:
-        for e in self.sub_exprs:
-            self.variables.update(e.variables)
+    for e in self.sub_exprs:
+        self.variables.update(e.variables)
     return self
 Expression.annotate1 = annotate1
 
@@ -710,7 +707,7 @@ def annotate(self, voc, q_vars):
                                 AComparison.make(op,
                                     deepcopy([applied, self.sub_exprs[0]])))
                     coc = AND([coc1, coc2])
-                    quantees = [Quantee.make(v, v.sort) for v in q_vars.values()]
+                    quantees = [Quantee.make(v, sort=v.sort) for v in q_vars.values()]
                     applied.co_constraint = FORALL(quantees, coc).annotate(voc, q_vars)
                     applied.co_constraint.annotations['reading'] = f"Calculation of {self.code}"
                 return applied
