@@ -38,7 +38,7 @@ if TYPE_CHECKING:
     from .Parse import Declaration, SymbolDeclaration, SymbolInterpretation, Enumeration
 
 from .utils import unquote, OrderedSet, BOOL, INT, REAL, DATE, CONCEPT, RESERVED_SYMBOLS, \
-    IDPZ3Error, DEF_SEMANTICS, Semantics
+    IDPZ3Error, Semantics
 
 
 class ASTNode(object):
@@ -165,7 +165,7 @@ class Constructor(ASTNode):
 
     def __str__(self):
         return (self.name if not self.sorts else
-                f"{self.name}({','.join((str(a) for a in self.sorts))}" )
+                f"{self.name}({', '.join((str(a) for a in self.sorts))}" )
 
 def CONSTRUCTOR(name: Symbol, args=None) -> Constructor:
     return Constructor(None, name, args)
@@ -222,12 +222,6 @@ class Expression(ASTNode):
             This is useful for definitions over infinite domains,
             as well as to compute relevant questions.
 
-        simpler (Expression, optional):
-            A simpler, equivalent expression.
-
-            Equivalence is computed in the context of the theory and structure.
-            Simplifying an expression is useful for efficiency
-            and to compute relevant questions.
 
         value (Optional[Expression]):
             A rigid term equivalent to the expression, obtained by
@@ -258,7 +252,6 @@ class Expression(ASTNode):
 
     def __init__(self):
         self.sub_exprs: List[Expression]
-        self.simpler: Optional[Expression] = None
         self.value: Optional[Expression] = None
 
         self.code: str = intern(str(self))
@@ -289,7 +282,6 @@ class Expression(ASTNode):
         out = copy(self)
         out.sub_exprs = [deepcopy(e, memo) for e in out.sub_exprs]
         out.variables = deepcopy(out.variables, memo)
-        out.simpler = None if out.simpler is None else deepcopy(out.simpler, memo)
         out.co_constraint = (None if out.co_constraint is None
                              else deepcopy(out.co_constraint, memo))
         if hasattr(self, 'questions'):
@@ -298,35 +290,40 @@ class Expression(ASTNode):
         return out
 
     def same_as(self, other: Expression):
-        if self.str == other.str:
+        # symmetric
+        if self.str == other.str: # and type(self) == type(other):
             return True
         if self.__class__.__name__ == "Number" and other.__class__.__name__ == "Number":
             return float(self.py_value) == float(other.py_value)
+
+        # asymetric
         if self.value is not None and self.value is not self:
-            return self.value  .same_as(other)
-        if self.simpler is not None:
-            return self.simpler.same_as(other)
-        if other.value is not None and other.value is not other:
-            return self.same_as(other.value)
-        if other.simpler is not None:
-            return self.same_as(other.simpler)
-
+            return self.value.same_as(other)
         if (isinstance(self, Brackets)
-           or (isinstance(self, AQuantification) and len(self.quantees) == 0)):
+           or (isinstance(self, AQuantification)
+               and len(self.quantees) == 0
+               and len(self.sub_exprs) == 1)):
             return self.sub_exprs[0].same_as(other)
-        if (isinstance(other, Brackets)
-           or (isinstance(other, AQuantification) and len(other.quantees) == 0)):
-            return self.same_as(other.sub_exprs[0])
 
-        return self.str == other.str and type(self) == type(other)
+        # switch role to be allowed to copy code
+        self, other = other, self
+
+        # copied code
+        if self.value is not None and self.value is not self:
+            return self.value.same_as(other)
+        if (isinstance(self, Brackets)
+           or (isinstance(self, AQuantification)
+               and len(self.quantees) == 0
+               and len(self.sub_exprs) == 1)):
+            return self.sub_exprs[0].same_as(other)
+
+        return False
 
     def __repr__(self): return str(self)
 
     def __str__(self):
         if self.value is not None and self.value is not self:
             return str(self.value)
-        if self.simpler is not None:
-            return str(self.simpler)
         return self.__str1__()
 
     def __log__(self):  # for debugWithYamlLog
@@ -508,7 +505,8 @@ class Expression(ASTNode):
                           level_symbols: Dict[SymbolDeclaration, Symbol],
                           head: AppliedSymbol,
                           pos_justification: bool,
-                          polarity: bool
+                          polarity: bool,
+                          mode: Semantics
                           ) -> Expression:
         """Returns an expression where level mapping atoms (e.g., lvl_p > lvl_q)
          are added to atoms containing recursive symbols.
@@ -524,7 +522,7 @@ class Expression(ASTNode):
             - polarity (Bool): whether the current expression occurs under
               negation.
         """
-        return (self.update_exprs((e.add_level_mapping(level_symbols, head, pos_justification, polarity)
+        return (self.update_exprs((e.add_level_mapping(level_symbols, head, pos_justification, polarity, mode)
                                    for e in self.sub_exprs))
                     .annotate1())  # update .variables
 
@@ -700,7 +698,10 @@ class Quantee(Expression):
         decl (SymbolDeclaration, Optional): the (unqualified) Declaration to quantify over, after resolution of `$(i)`.
         e.g., the declaration of `Color`
     """
-    def __init__(self, parent, vars, subtype=None, sort=None):
+    def __init__(self, parent,
+                 vars: List[List[Variable]],
+                 subtype: Type = None,
+                 sort: SymbolExpr = None):
         self.vars = vars
         self.subtype = subtype
         sort = sort
@@ -730,9 +731,10 @@ class Quantee(Expression):
     @classmethod
     def make(cls,
              var: List[Variable],
-             sort: SymbolExpr
+             subtype: Type = None,
+             sort: SymbolExpr = None
              ) -> 'Quantee':
-        out = (cls) (None, [var], sort)
+        out = (cls) (None, [var], subtype=subtype, sort=sort)
         return out.annotate1()
 
     def __str1__(self):
@@ -768,13 +770,15 @@ class AQuantification(Expression):
         self.quantees = quantees
         self.f = f
 
-        self.q = '∀' if self.q == '!' else '∃' if self.q == "?" else self.q
+        self.q = ('∀' if self.q in ['!', 'forall'] else
+                  '∃' if self.q in ["?", "thereisa"] else
+                  self.q)
         if self.quantees and not self.quantees[-1].sub_exprs:
             # separate untyped variables, so that they can be typed separately
             q = self.quantees.pop()
             for vars in q.vars:
                 for var in vars:
-                    self.quantees.append(Quantee.make(var, None))
+                    self.quantees.append(Quantee.make(var, sort=None))
 
         self.sub_exprs = [self.f]
         super().__init__()
@@ -793,11 +797,19 @@ class AQuantification(Expression):
         return out.annotate1()
 
     def __str1__(self):
-        if self.quantees:  #TODO this is not correct in case of partial expansion
-            vars = ','.join([f"{q}" for q in self.quantees])
-            return f"{self.q} {vars}: {self.sub_exprs[0].str}"
+        if len(self.sub_exprs) == 0:
+            body = TRUE.str if self.q == '∀' else FALSE.str
+        elif len(self.sub_exprs) == 1:
+            body = self.sub_exprs[0].str
         else:
-            return self.sub_exprs[0].str
+            connective = '∧' if self.q == '∀' else '∨'
+            body = connective.join("("+e.str+")" for e in self.sub_exprs)
+
+        if self.quantees:
+            vars = ','.join([f"{q}" for q in self.quantees])
+            return f"{self.q} {vars}: {body}"
+        else:
+            return body
 
     def __deecopy__(self, memo):
         # also called by AAgregate
@@ -826,15 +838,36 @@ def EXISTS(qs, expr, annotations=None):
 class Operator(Expression):
     PRECEDENCE = 0  # monkey-patched
     MAP = dict()  # monkey-patched
+    NORMAL = {
+        "is strictly less than": "<",
+        "is less than": "≤",
+        "=<": "≤",
+        "is greater than": "≥",
+        "is strictly greater than": ">",
+        ">=" : "≥",
+        "is not": "≠",
+        "~=": "≠",
+        "<=>": "⇔",
+        "is the same as": "⇔",
+        "<=": "⇐",
+        "are necessary conditions for": "⇐",
+        "=>": "⇒",
+        "then": "⇒",
+        "are sufficient conditions for": "⇒",
+        "|": "∨",
+        "or": "∨",
+        "&": "∧",
+        "and": "∧",
+        "*": "⨯",
+        "is": "=",
+    }
 
     def __init__(self, parent, operator, sub_exprs, annotations=None):
         self.operator = operator
         self.sub_exprs = sub_exprs
 
         self.operator = list(map(
-            lambda op: "≤" if op == "=<" else "≥" if op == ">=" else "≠" if op == "~=" else \
-                "⇔" if op == "<=>" else "⇐" if op == "<=" else "⇒" if op == "=>" else \
-                "∨" if op == "|" else "∧" if op == "&" else "⨯" if op == "*" else op
+            lambda op: Operator.NORMAL.get(op, op)
             , self.operator))
 
         super().__init__()
@@ -890,9 +923,9 @@ class Operator(Expression):
 class AImplication(Operator):
     PRECEDENCE = 50
 
-    def add_level_mapping(self, level_symbols, head, pos_justification, polarity):
-        sub_exprs = [self.sub_exprs[0].add_level_mapping(level_symbols, head, pos_justification, not polarity),
-                     self.sub_exprs[1].add_level_mapping(level_symbols, head, pos_justification, polarity)]
+    def add_level_mapping(self, level_symbols, head, pos_justification, polarity, mode):
+        sub_exprs = [self.sub_exprs[0].add_level_mapping(level_symbols, head, pos_justification, not polarity, mode),
+                     self.sub_exprs[1].add_level_mapping(level_symbols, head, pos_justification, polarity, mode)]
         return self.update_exprs(sub_exprs).annotate1()
 
 def IMPLIES(exprs, annotations=None):
@@ -917,9 +950,9 @@ def EQUIV(exprs, annotations=None):
 class ARImplication(Operator):
     PRECEDENCE = 30
 
-    def add_level_mapping(self, level_symbols, head, pos_justification, polarity):
-        sub_exprs = [self.sub_exprs[0].add_level_mapping(level_symbols, head, pos_justification, polarity),
-                     self.sub_exprs[1].add_level_mapping(level_symbols, head, pos_justification, not polarity)]
+    def add_level_mapping(self, level_symbols, head, pos_justification, polarity, mode):
+        sub_exprs = [self.sub_exprs[0].add_level_mapping(level_symbols, head, pos_justification, polarity, mode),
+                     self.sub_exprs[1].add_level_mapping(level_symbols, head, pos_justification, not polarity, mode)]
         return self.update_exprs(sub_exprs).annotate1()
 
 def RIMPLIES(exprs, annotations):
@@ -978,7 +1011,7 @@ class AUnary(Expression):
                  f: Expression):
         self.operators = operators
         self.f = f
-        self.operators = ['¬' if c == '~' else c for c in self.operators]
+        self.operators = ['¬' if c in ['~', 'not'] else c for c in self.operators]
         self.operator = self.operators[0]
         self.check(all([c == self.operator for c in self.operators]),
                    "Incorrect mix of unary operators")
@@ -994,11 +1027,12 @@ class AUnary(Expression):
     def __str1__(self):
         return f"{self.operator}({self.sub_exprs[0].str})"
 
-    def add_level_mapping(self, level_symbols, head, pos_justification, polarity):
+    def add_level_mapping(self, level_symbols, head, pos_justification, polarity, mode):
         sub_exprs = (e.add_level_mapping(level_symbols, head,
                                          pos_justification,
                                          not polarity
-                                         if self.operator == '¬' else polarity)
+                                         if self.operator == '¬' else polarity,
+                                         mode)
                      for e in self.sub_exprs)
         return self.update_exprs(sub_exprs).annotate1()
 
@@ -1023,7 +1057,9 @@ class AAggregate(Expression):
         super().__init__()
 
     def __str1__(self):
+        # aggregates are over finite domains, and cannot have partial expansion
         if not self.annotated:
+            assert len(self.sub_exprs) == 1, "Internal error"
             vars = ",".join([f"{q}" for q in self.quantees])
             if self.aggtype in ["sum", "min", "max"]:
                 out = (f"{self.aggtype}(lambda {vars} : "
@@ -1061,12 +1097,15 @@ class AppliedSymbol(Expression):
     Args:
         symbol (SymbolExpr): the symbol to be applied to arguments
 
-        is_enumerated (string): '' or 'is enumerated' or 'is not enumerated'
+        is_enumerated (string): '' or 'is enumerated'
 
-        is_enumeration (string): '' or 'in' or 'not in'
+        is_enumeration (string): '' or 'in'
 
         in_enumeration (Enumeration): the enumeration following 'in'
 
+        as_disjunction (Optional[Expression]):
+            the translation of 'is_enumerated' and 'in_enumeration' as a disjunction
+
         decl (Declaration): the declaration of the symbol, if known
 
         in_head (Bool): True if the AppliedSymbol occurs in the head of a rule
@@ -1091,6 +1130,7 @@ class AppliedSymbol(Expression):
 
         super().__init__()
 
+        self.as_disjunction = None
         self.decl = None
         self.in_head = False
 
@@ -1124,6 +1164,7 @@ class AppliedSymbol(Expression):
     def __deepcopy__(self, memo):
         out = super().__deepcopy__(memo)
         out.symbol = deepcopy(out.symbol)
+        out.as_disjunction = deepcopy(out.as_disjunction)
         return out
 
     def collect(self, questions, all_=True, co_constraints=True):
@@ -1153,9 +1194,8 @@ class AppliedSymbol(Expression):
         return symbols
 
     def has_decision(self):
-        self.check(self.decl.block is not None, "Internal error")
-        return not self.decl.block.name == 'environment' \
-            or any(e.has_decision() for e in self.sub_exprs)
+        return ((self.decl.block is not None and not self.decl.block.name == 'environment')
+            or any(e.has_decision() for e in self.sub_exprs))
 
     def type_inference(self):
         if self.symbol.decl:
@@ -1192,20 +1232,20 @@ class AppliedSymbol(Expression):
             constructor = CONSTRUCTOR(self.sub_exprs[0].name)
             constructors[symbol.name].append(constructor)
 
-    def add_level_mapping(self, level_symbols, head, pos_justification, polarity):
+    def add_level_mapping(self, level_symbols, head, pos_justification, polarity, mode):
         assert head.symbol.decl in level_symbols, \
                f"Internal error in level mapping: {self}"
         if self.symbol.decl not in level_symbols or self.in_head:
             return self
         else:
-            if DEF_SEMANTICS == Semantics.WELLFOUNDED:
+            if mode == Semantics.WELLFOUNDED:
                 op = ('>' if pos_justification else '≥') \
                     if polarity else ('≤' if pos_justification else '<')
-            elif DEF_SEMANTICS == Semantics.KRIPKEKLEENE:
+            elif mode == Semantics.KRIPKEKLEENE:
                 op = '>' if polarity else '≤'
             else:
-                assert DEF_SEMANTICS == Semantics.COINDUCTION, \
-                        f"Internal error: DEF_SEMANTICS"
+                assert mode == Semantics.COINDUCTION, \
+                        f"Internal error: {mode}"
                 op = ('≥' if pos_justification else '>') \
                     if polarity else ('<' if pos_justification else '≤')
             comp = AComparison.make(op, [
@@ -1277,17 +1317,18 @@ class Variable(Expression):
     Args:
         name (str): name of the variable
 
-        sort (Optional[Symbol]): sort of the variable, if known
+        sort (Optional[Union[Type, Symbol]]): sort of the variable, if known
     """
     PRECEDENCE = 200
 
     def __init__(self, parent,
                  name:str,
-                 sort: Optional[Symbol]=None):
+                 sort: Optional[Union[Type, Symbol]]=None):
         self.name = name
         sort = sort
         self.sort = sort
-        assert sort is None or isinstance(sort, Type) or isinstance(sort, Symbol)
+        assert sort is None or isinstance(sort, Type) or isinstance(sort, Symbol), \
+            f"Internal error: {self}"
 
         super().__init__()
 
@@ -1302,7 +1343,7 @@ class Variable(Expression):
 
     def annotate1(self): return self
 
-def VARIABLE(name, sort):
+def VARIABLE(name: str, sort: Union[Type, Symbol]):
     return Variable(None, name, sort)
 
 class Number(Expression):
