@@ -20,6 +20,7 @@
 Classes to parse an IDP-Z3 theory.
 
 """
+from __future__ import annotations
 
 from copy import copy, deepcopy
 from datetime import date
@@ -43,7 +44,8 @@ from .Expression import (Annotations, ASTNode, Constructor, CONSTRUCTOR, Accesso
     Variable, TRUEC, FALSEC, TRUE, FALSE, EQUALS, AND, OR, EQUIV)
 from .utils import (RESERVED_SYMBOLS, OrderedSet, NEWL, BOOL, INT, REAL, DATE, CONCEPT,
     GOAL_SYMBOL, EXPAND, RELEVANT, ABS, IDPZ3Error,
-    CO_CONSTR_RECURSION_DEPTH, MAX_QUANTIFIER_EXPANSION)
+    CO_CONSTR_RECURSION_DEPTH, MAX_QUANTIFIER_EXPANSION,
+    Semantics as S)
 
 
 def str_to_IDP(atom, val_string):
@@ -326,6 +328,8 @@ class TypeDeclaration(ASTNode):
         interpretation (SymbolInterpretation): the symbol interpretation
 
         map (Dict[string, Expression]): a mapping from code to Expression in range
+
+        voc (Vocabulary): the vocabulary block that contains it
     """
 
     def __init__(self, **kwargs):
@@ -351,9 +355,13 @@ class TypeDeclaration(ASTNode):
     def __str__(self):
         if self.name in RESERVED_SYMBOLS:
             return ''
-        enumeration = (f"{{{','.join(map(str, self.constructors))}}}" if self.constructors else
-                       f"{self.enumeration}" if self.enumeration else
-                       "")
+        enumeration = self.enumeration if hasattr(self, 'enumeration') and self.enumeration else ""
+        constructors = enumeration.constructors if enumeration else None
+        constructed = ("" if not bool(constructors) or all(0 == len(c.sorts) for c in constructors)
+                       else "constructed from ")
+        enumeration = (f"{constructed}{{{', '.join(str(c) for c in constructors)}}}" if constructors else
+                       f"{self.interpretation}" if self.interpretation else
+                       f"{enumeration}")
         return (f"type {self.name} {'' if not enumeration else ':= ' + enumeration}")
 
     def contains_element(self, term: Expression,
@@ -425,6 +433,8 @@ class SymbolDeclaration(ASTNode):
 
         needs_interpretation (bool):
             whether its interpretation must be sent to Z3
+
+        voc (Vocabulary): the vocabulary block that contains it
     """
 
     def __init__(self, **kwargs):
@@ -449,6 +459,7 @@ class SymbolDeclaration(ASTNode):
         self.unit: str = None
         self.heading: str = None
         self.optimizable: bool = True
+        self.needs_interpretation = False
 
         self.type = None  # a string
         self.base_type = None
@@ -603,12 +614,17 @@ class Definition(ASTNode):
     """
     definition_id = 0  # intentional static variable so that no two definitions get the same ID
 
-    def __init__(self, **kwargs):
+    def __init__(self, parent, annotations, mode, rules):
         Definition.definition_id += 1
         self.id = Definition.definition_id
-        self.annotations = kwargs.pop('annotations')
-        self.annotations = self.annotations.annotations if self.annotations else {}
-        self.rules = kwargs.pop('rules')
+        self.mode = (S.WELLFOUNDED if mode is None or 'well-founded' in mode else
+                     S.COMPLETION if 'completion' in mode else
+                     S.KRIPKEKLEENE if 'Kripke-Kleene' in mode else
+                     S.COINDUCTION if 'co-induction' in mode else
+                     mode)
+        assert type(self.mode) == S, f"Unsupported mode: {mode}"
+        self.annotations = annotations.annotations if annotations else {}
+        self.rules = rules
         self.clarks = {}  # {SymbolDeclaration: Transformed Rule}
         self.canonicals = {}
         self.instantiables = {}
@@ -705,10 +721,11 @@ class Rule(ASTNode):
         self.definiendum = kwargs.pop('definiendum')
         self.out = kwargs.pop('out')
         self.body = kwargs.pop('body')
-        self.is_whole_domain = None  # Bool
+        self.has_finite_domain = None  # Bool
         self.block = None  # theory where it occurs
 
-        self.annotations = self.annotations.annotations if self.annotations else {}
+        self.annotations = (self.annotations.annotations if self.annotations else
+                            {'reading': str(self)})
 
         if self.out is not None:
             self.definiendum.sub_exprs.append(self.out)
@@ -716,8 +733,9 @@ class Rule(ASTNode):
             self.body = TRUE
 
     def __repr__(self):
-        return (f"∀ {','.join(str(q) for q in self.quantees)}: "
-                f"{self.definiendum} "
+        quant = ('' if not self.quantees else
+                 f"∀ {','.join(str(q) for q in self.quantees)}: ")
+        return (f"{quant}{self.definiendum} "
                 f"{(' = ' + str(self.out)) if self.out else ''}"
                 f"← {str(self.body)}")
 
@@ -732,7 +750,6 @@ class Rule(ASTNode):
             Expression: a boolean expression
         """
 
-        #TODO assert self.is_whole_domain == False
         out = deepcopy(self.body)  # in case there are no arguments
         instance = AppliedSymbol.make(self.definiendum.symbol, new_args)
         instance.in_head = True
@@ -813,6 +830,9 @@ class SymbolInterpretation(ASTNode):
         self.symbol = None
         self.is_type_enumeration = None
 
+    def __repr__(self):
+        return f"{self.name} {self.sign} {self.enumeration}"
+
     def interpret_application(self, rank, applied, args, tuples=None):
         """returns an expression equivalent to `self.symbol` applied to `args`,
         simplified by the interpretation of `self.symbol`.
@@ -846,8 +866,9 @@ class SymbolInterpretation(ASTNode):
             key = ",".join(a.code for a in args)
             if key in self.enumeration.lookup:
                 return self.enumeration.lookup[key]
-        # in case of partial interpretation
-        return applied._change(sub_exprs=args)
+            elif self.enumeration.parent.sign != '≜':
+                return applied._change(sub_exprs=args)
+            return self.enumeration.parent.default
     """
         if rank == self.symbol.decl.arity:  # valid tuple -> return a value
             if not type(self.enumeration) == FunctionEnum:
@@ -1039,7 +1060,7 @@ class TupleIDP(ASTNode):
         return self.code
 
     def __repr__(self):
-        return self.code
+        return f"({self.code})" if 1 < len(self.args) else self.code
 
 
 class FunctionTuple(TupleIDP):
@@ -1118,7 +1139,7 @@ class Ranges(Enumeration):
                   ) -> Extension:
         if not self.elements:
             return(None, None)
-        if self.tuples: # and len(self.tuples) < MAX_QUANTIFIER_EXPANSION:
+        if self.tuples is not None: # and len(self.tuples) < MAX_QUANTIFIER_EXPANSION:
             return ([t.args for t in self.tuples], None)
         def filter(args):
             sub_exprs = []
@@ -1132,6 +1153,12 @@ class Ranges(Enumeration):
         return(None, filter)
 
 
+class RangeElement(ASTNode):
+    def __init__(self, **kwargs):
+        self.fromI = kwargs.pop('fromI')
+        self.toI = kwargs.pop('toI')
+
+
 class IntRange(Ranges):
     def __init__(self):
         Ranges.__init__(self, elements=[])
@@ -1346,7 +1373,7 @@ idpparser = metamodel_from_file(dslFile, memoization=True,
                                          Enumeration, FunctionEnum, CSVEnumeration,
                                          TupleIDP, FunctionTuple, CSVTuple,
                                          ConstructedFrom, Constructor, Ranges,
-                                         Display,
+                                         RangeElement, Display,
 
                                          Procedure, Call1, String,
                                          PyList, PyAssignment])
