@@ -27,6 +27,7 @@ It has 2 parts:
 
 This module monkey-patches the Expression and Theory classes and sub-classes.
 """
+from __future__ import annotations
 
 import time
 from copy import copy
@@ -61,8 +62,6 @@ def simplify_with(self: Expression, assignments: "Assignments") -> Expression:
     if self.value is not None:
         return self
     value, simpler, new_e, co_constraint = None, None, None, None
-    if self.simpler is not None:
-        simpler = self.simpler.simplify_with(assignments)
     if self.co_constraint is not None:
         co_constraint = self.co_constraint.simplify_with(assignments)
     new_e = [e.simplify_with(assignments) for e in self.sub_exprs]
@@ -97,10 +96,7 @@ def symbolic_propagate(self,
     if self.value is None:
         if self.code in assignments:
             assignments.assert__(self, truth, tag)
-        if self.simpler is not None:
-            self.simpler.symbolic_propagate(assignments, tag, truth)
-        else:
-            self.propagate1(assignments, tag, truth)
+        self.propagate1(assignments, tag, truth)
 Expression.symbolic_propagate = symbolic_propagate
 
 
@@ -115,10 +111,11 @@ Expression.propagate1 = propagate1
 def symbolic_propagate(self, assignments, tag, truth=TRUE):
     if self.code in assignments:
         assignments.assert__(self, truth, tag)
-    if not self.quantees:  # expanded
-        assert len(self.sub_exprs) == 1,  \
-               f"Internal error in symbolic_propagate: {self}"  # a conjunction or disjunction
-        self.sub_exprs[0].symbolic_propagate(assignments, tag, truth)
+    if not self.quantees:  # expanded, no variables  dead code ?
+        if self.q == '∀' and truth.same_as(TRUE):
+            AConjunction.symbolic_propagate(self, assignments, tag, truth)
+        elif truth.same_as(FALSE):
+            ADisjunction.symbolic_propagate(self, assignments, tag, truth)
 AQuantification.symbolic_propagate = symbolic_propagate
 
 
@@ -148,6 +145,15 @@ def propagate1(self, assignments, tag, truth=TRUE):
 AUnary.propagate1 = propagate1
 
 
+# class AppliedSymbol  ############################################################
+
+def propagate1(self, assignments, tag, truth=TRUE):
+    if self.as_disjunction:
+        self.as_disjunction.symbolic_propagate(assignments, tag, truth)
+    Expression.propagate1(self, assignments, tag, truth)
+AUnary.propagate1 = propagate1
+
+
 # class AComparison  ##########################################################
 
 def propagate1(self, assignments, tag, truth=TRUE):
@@ -179,11 +185,11 @@ Brackets.symbolic_propagate = symbolic_propagate
 #
 ###############################################################################
 
-def _set_consequences_get_changed_choices(self):
+def _set_consequences_get_changed_choices(self, tag):
     # clear consequences, as these might not be cleared by add_given when
     # running via CLI
     for a in self.assignments.values():
-        if a.status in [S.CONSEQUENCE, S.ENV_CONSQ]:
+        if a.status == tag:
             self.assignments.assert__(a.sentence, None, S.UNKNOWN)
 
     removed_choices = {a.sentence.code: a for a in self.previous_assignments.values()
@@ -200,7 +206,7 @@ def _set_consequences_get_changed_choices(self):
 
     if not removed_choices:
         for a in self.previous_assignments.values():
-            if (a.status in [S.CONSEQUENCE, S.ENV_CONSQ] and
+            if (a.status == tag and
                     self.assignments[a.sentence.code].status
                     not in [S.GIVEN, S.EXPANDED, S.DEFAULT]):
                 self.assignments.assert__(a.sentence, a.value, a.status)
@@ -412,14 +418,15 @@ def _propagate(self, tag=S.CONSEQUENCE, given_todo=None):
         yield from self._propagate_ignored(tag, given_todo)
         return
 
+    solver = self.solver
     if not self.previous_assignments:
         try:
-            yield from self._first_propagate(self.solver)
+            yield from self._first_propagate(solver)
         except IDPZ3Error:
             yield NOT_SATISFIABLE
             # can't access solver.sexpr()
 
-    removed_choices, added_choices = self._set_consequences_get_changed_choices()
+    removed_choices, added_choices = self._set_consequences_get_changed_choices(tag)
 
     dir_todo = given_todo is None
     if dir_todo:
@@ -438,11 +445,6 @@ def _propagate(self, tag=S.CONSEQUENCE, given_todo=None):
         for a in to_remove:
             todo.pop(a.code)
 
-    if not todo:
-        yield "No more consequences."
-        return
-
-    solver = self.solver
     solver.push()
 
     self._add_assignment(solver)
