@@ -18,12 +18,13 @@
 """
 This module contains the logic for the computation of relevance.
 """
+from __future__ import annotations
 
-from idp_engine.Assignments import Status as S
-from idp_engine.Expression import (AppliedSymbol, TRUE, Expression, AQuantification,
+from .Assignments import Status as S
+from .Expression import (AppliedSymbol, TRUE, Expression, AQuantification,
                                    AConjunction, Brackets, UnappliedSymbol)
-from idp_engine.Theory import Theory
-from idp_engine.utils import OrderedSet, GOAL_SYMBOL, RELEVANT
+from .Theory import Theory
+from .utils import OrderedSet, GOAL_SYMBOL, RELEVANT
 
 
 def split_constraints(constraints: OrderedSet) -> OrderedSet:
@@ -44,12 +45,9 @@ def split_constraints(constraints: OrderedSet) -> OrderedSet:
             for e in c.sub_exprs:
                 split(e, cs)
         elif type(c) == AQuantification and c.q == '∀':
-            assert len(c.sub_exprs) == 1, "Internal error in split"
             conj = OrderedSet()
-            if c.simpler:
-                split(c.simpler, conj)
-            else:
-                split(c.sub_exprs[0], conj)
+            for e in c.sub_exprs:
+                split(e, conj)
             for e in conj:
                 out = AQuantification.make(c.q, c.quantees, e)
                 # out.code = c.code
