@@ -23,15 +23,16 @@ This module monkey-patches the Expression class and sub-classes.
 
 
 """
+from __future__ import annotations
 
 from copy import deepcopy
 import sys
 from typing import List
 
 from .Expression import (
-    Constructor, Expression, AIfExpr, AQuantification, Quantee, Type,
+    Constructor, Expression, AIfExpr, IF, AQuantification, Quantee, Type,
     Operator, AEquivalence, AImplication, ADisjunction,
-    AConjunction, AComparison, ASumMinus, AMultDiv, APower,
+    AConjunction, AComparison, EQUALS, ASumMinus, AMultDiv, APower,
     AUnary, AAggregate, SymbolExpr, AppliedSymbol, UnappliedSymbol, Variable,
     Number, Date, Brackets, TRUE, FALSE, NOT, AND, OR)
 from .Parse import Symbol, Enumeration, TupleIDP, TypeDeclaration
@@ -57,7 +58,13 @@ def _change(self, sub_exprs=None, ops=None, value=None, simpler=None,
         if simpler.value is not None:  # example: prime.idp
             self.value = simpler.value
         else:
-            self.simpler = simpler
+            simpler.annotations = self.annotations
+            #simpler.code = self.code
+            simpler.original = self.original
+            simpler.is_type_constraint_for = self.is_type_constraint_for
+            simpler.co_constraint = self.co_constraint
+            simpler.block = self.block if hasattr(self, "block") else None
+            return simpler
     assert (self.value is None
                or type(self.value) in [AppliedSymbol, UnappliedSymbol, Symbol,
                                        Number, Date, Type]), \
@@ -85,7 +92,6 @@ Expression.simplify1 = simplify1
 
 # Class AIfExpr  ###############################################################
 
-
 def update_exprs(self, new_exprs):
     sub_exprs = list(new_exprs)
     if_, then_, else_ = sub_exprs[0], sub_exprs[1], sub_exprs[2]
@@ -96,11 +102,16 @@ def update_exprs(self, new_exprs):
     else:
         if then_.same_as(else_):
             return self._change(simpler=then_, sub_exprs=sub_exprs)
-        elif then_.same_as(TRUE) and else_.same_as(FALSE):
-            return self._change(simpler=if_, sub_exprs=sub_exprs)
-        elif then_.same_as(FALSE) and else_.same_as(TRUE):
-            return self._change(simpler=NOT(if_),
-                                sub_exprs=sub_exprs)
+        elif then_.same_as(TRUE):
+            if else_.same_as(FALSE):
+                return self._change(simpler=if_, sub_exprs=sub_exprs)
+            else:
+                return self._change(simpler=OR([if_, else_]), sub_exprs=sub_exprs)
+        elif else_.same_as(TRUE):
+            if then_.same_as(FALSE):
+                return self._change(simpler=NOT(if_), sub_exprs=sub_exprs)
+            else:
+                return self._change(simpler=OR([NOT(if_), then_]), sub_exprs=sub_exprs)
     return self._change(sub_exprs=sub_exprs)
 AIfExpr.update_exprs = update_exprs
 
@@ -120,14 +131,10 @@ Quantee.update_exprs = update_exprs
 # Class AQuantification  ######################################################
 
 def update_exprs(self, new_exprs):
-    exprs = list(new_exprs)
-    if not self.quantees:
-        if self.q == '∀':
-            simpler = AND(exprs)
-        else:
-            simpler = OR(exprs)
-        return self._change(simpler=simpler, sub_exprs=[simpler])
-    return self._change(sub_exprs=exprs)
+    if self.q == '∀':
+        return AConjunction.update_exprs(self, new_exprs, replace=False)
+    else:
+        return ADisjunction.update_exprs(self, new_exprs, replace=False)
 AQuantification.update_exprs = update_exprs
 
 
@@ -177,7 +184,7 @@ AEquivalence.update_exprs = update_exprs
 
 # Class ADisjunction  #######################################################
 
-def update_exprs(self, new_exprs):
+def update_exprs(self, new_exprs, replace=True):
     exprs, other = [], []
     value, simpler = None, None
     for i, expr in enumerate(new_exprs):
@@ -192,7 +199,7 @@ def update_exprs(self, new_exprs):
 
     if len(other) == 0:  # all disjuncts are False
         value = FALSE
-    if len(other) == 1:
+    if replace and len(other) == 1:
         simpler = other[0]
     return self._change(value=value, simpler=simpler, sub_exprs=exprs)
 ADisjunction.update_exprs = update_exprs
@@ -201,7 +208,7 @@ ADisjunction.update_exprs = update_exprs
 # Class AConjunction  #######################################################
 
 # same as ADisjunction, with TRUE and FALSE swapped
-def update_exprs(self, new_exprs):
+def update_exprs(self, new_exprs, replace=True):
     exprs, other = [], []
     value, simpler = None, None
     for i, expr in enumerate(new_exprs):
@@ -216,7 +223,7 @@ def update_exprs(self, new_exprs):
 
     if len(other) == 0:  # all conjuncts are True
         value = TRUE
-    if len(other) == 1:
+    if replace and len(other) == 1:
         simpler = other[0]
     return self._change(value=value, simpler=simpler, sub_exprs=exprs)
 AConjunction.update_exprs = update_exprs
@@ -226,6 +233,19 @@ AConjunction.update_exprs = update_exprs
 
 def update_exprs(self, new_exprs):
     operands = list(new_exprs)
+
+    if len(operands) == 2 and self.operator == ["="]:
+        # a = a
+        if operands[0].same_as(operands[1]):
+            return self._change(value=TRUE, sub_exprs=operands)
+
+        # (if c then a else b) = d  ->  (if c then a=d else b=d)
+        if type(operands[0]) == AIfExpr:
+            then = EQUALS([operands[0].sub_exprs[1], operands[1]]).simplify1()
+            else_ = EQUALS([operands[0].sub_exprs[2], operands[1]]).simplify1()
+            new = IF(operands[0].sub_exprs[0], then, else_).simplify1()
+            return self._change(simpler=new, sub_exprs=operands)
+
     operands1 = [e.value for e in operands]
     acc, acc1 = operands[0], operands1[0]
     assert len(self.operator) == len(operands1[1:]), "Internal error"
@@ -237,6 +257,9 @@ def update_exprs(self, new_exprs):
             if op == "=":
                 if not acc1.same_as(expr1):
                     return self._change(value=FALSE, sub_exprs=[acc, expr], ops=[op])
+            if op == "≠":  # issue #246
+                if acc1.same_as(expr1):
+                    return self._change(value=FALSE, sub_exprs=[acc, expr], ops=[op])
             elif not (Operator.MAP[op]) (acc1.py_value, expr1.py_value):
                 return self._change(value=FALSE, sub_exprs=[acc, expr], ops=[op])
         acc, acc1 = expr, expr1
@@ -456,7 +479,7 @@ def join_set_conditions(assignments: List[Assignment]) -> List[Assignment]:
                     )
 
                     core = deepcopy(AppliedSymbol.make(out.symbol, out.sub_exprs))
-                    out.simpler = out.in_enumeration.contains([core], False)
+                    out.as_disjunction = out.in_enumeration.contains([core], False)
 
                     out = Assignment(out,
                                      TRUE if belongs else FALSE,
