@@ -18,6 +18,7 @@
 """
     Various utilities (in particular, OrderedSet)
 """
+from __future__ import annotations
 
 from collections import ChainMap
 from collections.abc import Iterable
@@ -43,8 +44,6 @@ class Semantics(Enum):
     WELLFOUNDED = auto()
     COINDUCTION = auto()
 
-DEF_SEMANTICS = Semantics.WELLFOUNDED
-
 """
     String constants
 """
@@ -159,3 +158,43 @@ class OrderedSet(dict):
         """
         out = OrderedSet({v for v in self if v not in other})
         return out
+
+
+# redirection #############################################
+# https://stackoverflow.com/a/22434728/474491
+
+import os
+import sys
+from contextlib import contextmanager
+
+def fileno(file_or_fd):
+    fd = getattr(file_or_fd, 'fileno', lambda: file_or_fd)()
+    if not isinstance(fd, int):
+        raise ValueError("Expected a file (`.fileno()`) or a file descriptor")
+    return fd
+
+@contextmanager
+def redirect_stdout(to=os.devnull, stdout=None):
+    if stdout is None:
+       stdout = sys.stdout
+
+    stdout_fd = fileno(stdout)
+    # copy stdout_fd before it is overwritten
+    #NOTE: `copied` is inheritable on Windows when duplicating a standard stream
+    with os.fdopen(os.dup(stdout_fd), 'wb') as copied:
+        stdout.flush()  # flush library buffers that dup2 knows nothing about
+        try:
+            os.dup2(fileno(to), stdout_fd)  # $ exec >&to
+        except ValueError:  # filename
+            with open(to, 'wb') as to_file:
+                os.dup2(to_file.fileno(), stdout_fd)  # $ exec > to
+        try:
+            yield stdout # allow code to be run with the redirected stdout
+        finally:
+            # restore stdout to its previous value
+            #NOTE: dup2 makes stdout_fd inheritable unconditionally
+            stdout.flush()
+            os.dup2(copied.fileno(), stdout_fd)  # $ exec >&copied
+
+def redirect_stderr_to_stdout():  # $ exec 2>&1
+    return redirect_stdout(to=sys.stdout, stdout=sys.stderr)
\ No newline at end of file
